/* copyright-start
  ******************************************************************************
  *  Licensed Material - Property of IBM
  *
  *  5724-V27, 5655-R10, 5655-M22, 5655-L21, 5655-I49, 5655-D92
  *  Copyright IBM Corp. 1993, 2009                          
  *
  *  The source code for this program is not published or otherwise
  *  divested of its trade secrets, irrespective of what has been
  *  deposited with the U.S. Copyright Office.
  ******************************************************************************
  * copyright-end */

package com.ibm.dmh.ram;

import java.io.IOException;
import java.net.Authenticator;
import java.net.HttpURLConnection;
import java.net.PasswordAuthentication;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.helpers.NullEnumeration;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.jdom.Element;

import com.ibm.dmh.rest.client.DmhArtifact;
import com.ibm.ram.client.RAMAction;
import com.ibm.ram.client.RAMAsset;
import com.ibm.ram.client.RAMAssetAttribute;
import com.ibm.ram.client.RAMAssetQueryBuilder;
import com.ibm.ram.client.RAMAssetType;
import com.ibm.ram.client.RAMAssetValidation;
import com.ibm.ram.client.RAMAttributeConstraint;
import com.ibm.ram.client.RAMCategory;
import com.ibm.ram.client.RAMCategorySchema;
import com.ibm.ram.client.RAMCommunity;
import com.ibm.ram.client.RAMFolderArtifact;
import com.ibm.ram.client.RAMRelationshipType;
import com.ibm.ram.client.RAMSession;
import com.ibm.ram.client.RAMURLArtifact;
import com.ibm.ram.client.RAMUser;
import com.ibm.ram.common.data.Artifact;
import com.ibm.ram.common.data.AssetAttribute;
import com.ibm.ram.common.data.CategorySchema;
import com.ibm.ram.common.data.Relationship;
import com.ibm.ram.common.data.SearchResult;

/**
 * @author pboney@us.ibm.com, Leshek@us.ibm.com
 *
 * This class intereacts directly with a RAM instance.  Used for integrating RAA into RAM
 */
public class DmhRam {

	private String raaUser 	   = null;
	private String raaPassword = null;
	private String ramUri      = null;
	private String ramUser     = null;
	private String ramPassword = null;
	private RAMSession ramSession = null;
	private Element raaAttributesToPopulateInRam = null;

	private DmhRam() {
		throw new RuntimeException("DmhRam - default constructor forbiden - must provide uri, user, password");
	}

   /**
    * @param ramUri
    * @param ramUser
    * @param ramPassword
    */
   public DmhRam(String ramUri, String ramUser, String ramPassword) {
      super();
      this.ramUri = ramUri;
      this.ramUser = ramUser;
      this.ramPassword = ramPassword;
      // initialize logging if not already initialized
      if (Logger.getRootLogger().getAllAppenders() instanceof NullEnumeration) {
         BasicConfigurator.configure();
         // TRACE, DEBUG, INFO, WARN, ERROR and FATAL
         Logger.getRootLogger().setLevel(Level.ERROR);         
      }
   }

	/**
	 * maintain session to RAM 
	 * @return The session to acces the RAM instance
	 */ 
	public RAMSession getRAMSession() {
		if (ramSession == null) 
			ramSession = new RAMSession(ramUri, ramUser, ramPassword);
		return ramSession;
	}//end getRamSession
	
	/**
	 * Return existing or create a relationship type.  If it already exists, do nothing. Name the reverse relationship is needed. 
	 * @param relationshipDisplayName = the name that will be displayed when the relationship is added to an asset
	 * @return ramRelationshipName
	 */
	public RAMRelationshipType createRelationship(String relationshipDisplayName, String relationshipDisplayNameReverse){
		//	Fetch a relationship type by Display Name      
      String ramRelationshipName = DmhArtifact.makeRamRelationshipName(relationshipDisplayName);
      if (relationshipDisplayNameReverse == null || relationshipDisplayNameReverse.trim().equals("")) {
         relationshipDisplayNameReverse = ramRelationshipName+"-reverse";
      } else {
         relationshipDisplayNameReverse = DmhArtifact.makeRamRelationshipName(relationshipDisplayNameReverse);
      }
		RAMRelationshipType relationshipType = getRAMSession().getRelationshipType(ramRelationshipName);
		if (relationshipType == null){//create the relationship type
			RAMRelationshipType newRelationshipType = getRAMSession().createRelationshipType(ramRelationshipName, relationshipDisplayNameReverse);            
         getRAMSession().put(newRelationshipType, new NullProgressMonitor());
         //commitChanges();
		} 
      return getRAMSession().getRelationshipType(ramRelationshipName);
	}//end getCreateRelationship
	
	/**
	 * Creates a user in RAM with the given username and contact information.
	 * If the contact information is null, generic values are inserted
	 * @param userName = the user name to be uploaded to ram
	 * @param email = the user's email address
	 * @param phone = the user's telephone number
	 * @return RAMUser
	 */
	public RAMUser createUser(String userName, String email, String phone) {
		RAMUser user =  getRAMSession().getUser(userName);
        
		if(!user.isRegistered()){  //Register user
			if (email == null) email = "Not specified";
			if (phone == null) phone = "Not specified";
			user.setName(userName);
			user.setEmail(email);
			user.setPhone(phone);
			getRAMSession().put(user, new NullProgressMonitor()); 
		}
      return user;
	}//end createUser
	
	/**
	 * Takes a commnity name and description as parameters.  If the community exists, do nothing, else create it
	 * @param communityDesc = description of the comminuty to be displayed in RAM
	 * @return community
	 */
	 public RAMCommunity createCommunity(String communityName, String communityDesc) {
      // check for RAA Community. create it if necessary
      RAMCommunity community = getRAMSession().getCommunity(communityName);
      if (community == null) { // community not found - create
         community = getRAMSession().createCommunity(communityName);
         community.setDescription(communityDesc);
         getRAMSession().put(community, new NullProgressMonitor());
      }
      return community;
   }// end createCommunity
	 
	 /** 
	  * Takes an attribute name and description as a parameter.  If it does not exist, create the custom attribute.
	  * Then open the RAA Type and add the attribute as a non-required constraint
	  * @param attributeName = the desired attribute name
	  * @param attributeDesc = description of the attribute to be displayed in RAM
	  * @return attributeDetails
	  */
	 public RAMAssetAttribute createAttribute(String attributeName, String attributeDesc) {
      RAMAssetAttribute attributeDetails = getRAMSession().getAssetAttribute(attributeName);
      if (attributeDetails == null) {
         if (attributeDesc == null ) attributeDesc = attributeName;
         attributeDetails = getRAMSession().createAssetAttribute(attributeName);
         attributeDetails.setDescription(attributeDesc);
         getRAMSession().put(attributeDetails, new NullProgressMonitor());
         //commitChanges();
         
         //now fetch the RAA Type and add the attribute constraint
         RAMAssetType assetType = getRAMSession().getAssetType(DmhArtifact.getRamAssetType());
         if (assetType == null)
        	 System.out.println("BAD INPUT: RAMAssetType "+DmhArtifact.getRamAssetType()+" DOES NOT EXIST.  ATTRIBUTE NOT ADDED TO TYPE");
         else {                          //add the attribute as a constraint
            RAMAttributeConstraint[] ac = assetType.getAttributeConstraints(); //we are required to use this workaround of re-constraining attributes
            for (int i = 0; ac != null &&  i < ac.length; i++) {               
                assetType.createAttributeConstraint(ac[i].getAssetAttribute(), false);
            }
        	 assetType.createAttributeConstraint(attributeDetails, false);
        	 getRAMSession().put(assetType, new NullProgressMonitor());
         }
      }
      return attributeDetails;
   }// end createAttribute
	 
	 /**
	  * Create the RAA CategorySchema if it has not already been initialized
	  * @param schemaName = desired name of the schema
	  * @param schemaModelName = desired model name for the schema
	  * @param args = String array of subcategories to be placed under the model
	  * @return schema
	  */
	 public RAMCategorySchema createCategorySchema(String schemaName, String schemaModelName, String[] args) {
      RAMSession session = getRAMSession();
      RAMCategorySchema schema = session.getCategorySchema(schemaName);
      if (schema == null) {// create it
         RAMCategorySchema newCategorySchema = session.createCategorySchema(schemaName);
         // Create new Category Model
         RAMCategory newCategory = newCategorySchema.createCategory(schemaModelName);
         // Add model children - RAA artifact tpes
         for (int i = 0; args != null && i < args.length; i++)
            newCategory.createSubCategory(args[i]);

         session.put(newCategorySchema, new NullProgressMonitor());
         //commitChanges();
         schema = session.getCategorySchema(schemaName);
      }
      return schema;
	 }// end createCategorySchema
	 
	 public void addSubCategory(String subcategory){
		 //get the current Schema
		 RAMCategorySchema schema = getRAMSession().getCategorySchema(DmhArtifact.getRamCategorySchemaName());
         
         //Get the current category model
         RAMCategory category = (RAMCategory)schema.getCategory(DmhArtifact.getRamCategoryModelName());
         
         //Create a new subcategory
         category.createSubCategory(subcategory);

         getRAMSession().put(schema, new NullProgressMonitor());
         //commitChanges();
	 }//end addSubCategory
	 
     /**
      * Creates an asset with name typeName if it has not already been initialized
      * @param typeName = desired name for the type
      * @param typeDescription = description to be displayed in RAM
      * @return assetType
      */
	 public RAMAssetType createAssetType(String typeName, String typeDescription) {
		 if (typeDescription == null) typeDescription = typeName;      
		 RAMAssetType assetType = getRAMSession().getAssetType(typeName);
		 if (assetType == null) {
			 RAMAssetType type = getRAMSession().createAssetType(typeName);
			 type.setDescription(typeDescription);
			 
			 getRAMSession().put(type, new NullProgressMonitor());
			 putAssets();
			 assetType = getRAMSession().getAssetType(typeName);
		 }
		 return assetType;  
	 }// end createAssetType
	 	
	 /**
     * Sets up the RAM environment for use with RAA. This will create all the
     * necessary items in RAM for integration.
     */
    public void setRamEnvironment(String[] catArgs ){
    	//create RAA user
    	createUser("RAA", null, null);
    	
   	 //create community
    	createCommunity(DmhArtifact.getRamCommunityName(), "Community for all RAA artifacts.");
    	
   	 //create the category schema
   	 	createCategorySchema(DmhArtifact.getRamCategorySchemaName(), DmhArtifact.getRamCategoryModelName(), catArgs);
   	      
       //create the type
       createAssetType(DmhArtifact.getRamAssetType(), "Type for all RAA artifacts");
   
   	 //create the custom attributes and add them as non-required constraints to RAA Type
       	createAttribute(DmhArtifact.getRamRaaAttributesNameDetailsPage(), "Details page link for the RAA artifact.");
   	 	createAttribute(DmhArtifact.getRamRaaAttributesName(), "RAA attributes");
   	 	createAttribute(DmhArtifact.getRamRaaAttributesNameRaaHome(), "Originating RAA home page");
   	 	createAttribute(DmhArtifact.makeRamAttributeName("type"), DmhArtifact.makeRamAttributeDescription("type"));
    }//end setRamEnvironment

   /**
	 * Deletes an asset of the given UID
	 * @param uid = the RAM unique ID of the asset
	 */
	public void deleteAsset(String uid){
		RAMAsset asset = getRAMSession().getAsset(uid, DmhArtifact.getVersion());
		asset.setAction(RAMAction.DELETE);
		getRAMSession().put(asset, new NullProgressMonitor());
	}//end deleteAsset

	/** 
	 * This method deletes the RAA assets, community, type, category schema
	 */
	public void deleteRaaRamEnvironment() {
	  //DELETION OF USERS IS NOT SUPPORTED
      
      // delete the community and all assets within it
      RAMCommunity community = getRAMSession().getCommunity(DmhArtifact.getRamCommunityName());
      if (community != null) {
         community.setAction(RAMAction.DELETE);;
         getRAMSession().put(community, new NullProgressMonitor());
      }

      // delete the type
      RAMAssetType assetType = getRAMSession().getAssetType(DmhArtifact.getRamAssetType());
      if (assetType != null) {
         assetType.setAction(RAMAction.DELETE);;
         getRAMSession().put(assetType, new NullProgressMonitor());
      }

      // delete the category schema
      RAMCategorySchema categorySchema = getRAMSession().getCategorySchema(DmhArtifact.getRamCategorySchemaName());
      if (categorySchema != null) {
         categorySchema.setAction(RAMAction.DELETE);;
         getRAMSession().put(categorySchema, new NullProgressMonitor());
      }

      // Delete the "RAA *" attributes
      RAMAssetAttribute[] allAttrs = getRAMSession().getAllAssetAttributes();
      for (int i = 0; i < allAttrs.length; i++)
         if (allAttrs[i].getName().startsWith("RAA")) {
            allAttrs[i].setAction(RAMAction.DELETE);;
            getRAMSession().put(allAttrs[i], new NullProgressMonitor());
         }
      
      // delete the RAA relationship types
      RAMRelationshipType[] relationships = getRAMSession().getAllRelationshipTypes();
      for (int i = 0; i < relationships.length; i++)
         if (relationships[i].toString().startsWith("RAA") && !relationships[i].toString().endsWith("-reverse")) {
            relationships[i].setAction(RAMAction.DELETE);
            try{
            getRAMSession().put(relationships[i], new NullProgressMonitor());
            } catch (com.ibm.ram.common.data.exception.RAMRuntimeException e){}
              catch(NullPointerException n){}
         }
      
   }//end deleteRaaRamEnvironment
	
	/**
	 * Creates an asset based on a DmhArtifact.  If the asset is already in RAM, the asset is fetched then updated.
	 * and submits it to the server. 
	 * @param artifact = an artifact from RAA
	 * @return ramUid
	 */
	public RAMAsset uploadAsset(DmhArtifact artifact) {
		ramSession = null;
		//search by asset name to see if the asset already exists
      RAMAsset asset = getRamAssetByGuid(artifact.getRamGUID(), DmhArtifact.getVersion());
		if( asset != null ){	// the asset already exists. modify and update it
			setRamAssetProperties(artifact, asset);
			getRAMSession().put(asset, new NullProgressMonitor());
		}
		else {  	// Create a new asset to submit to the server
			asset = getRAMSession().createAsset(artifact.getRamGUID(),DmhArtifact.getVersion());
			setRamAssetProperties(artifact, asset);		
			categorizeAsset(asset, artifact.getCategoryName()); 	// Now categorize the asset
			try{getRAMSession().put(asset, new NullProgressMonitor());}
			catch(com.ibm.ram.common.data.exception.RAMRuntimeException e){e.printStackTrace();} //if asset already exists, do nothing.  (already handled by updating the asset)
		}	
		return asset;
	}//end uploadAsset

	/**
	 * sets properties for an asset before uploading it to RAM
	 * @param raaArtifact = artifact from RAA
	 * @param ramAsset = the RAM asset representing the RAA artifact
	 */
   private void setRamAssetProperties(DmhArtifact raaArtifact, RAMAsset ramAsset) {
		RAMSession session = getRAMSession();
		ramAsset.setName(raaArtifact.getRamName());
		ramAsset.setCommunity(session.getCommunity(DmhArtifact.getRamCommunityName()));
		ramAsset.setAssetType(session.getAssetType(DmhArtifact.getRamAssetType()));
		ramAsset.setShortDescription(raaArtifact.getRamShortDescription());
		ramAsset.setOwners(new RAMUser[] { session.getUser(raaArtifact.getRamOwner()) });
		
		// Set RAA attributes - set RAA attributes into RAA Attributes RAM attribute
        addGenericAttributesContent(raaArtifact, ramAsset);

        setRamAssetPropertiesCustomAttributes(raaArtifact, ramAsset);

		// Validate and report errors if any - mostly for testing 
		RAMAssetValidation[] messages = ramAsset.validate();
		String messagesText = "";
		for (int i = 0; ( messages != null && i < messages.length); i++) {
			System.out.println(messages[i].getMessage());
			messagesText = messagesText + messages[i].getMessage();
		}
		if (messages != null && messages.length > 0) {
         messagesText = "Setting properties for: "+ raaArtifact.getRaaDetailsPageLink() +"\n"+
                        "Validation failed for "+raaArtifact.getRamName()+": " + messagesText;
         Logger.getRootLogger().error(messagesText);
			throw new RuntimeException(messagesText);
		}
	}//end setRamAssetProperties

   private void addGenericAttributesContent(DmhArtifact raaArtifact, RAMAsset ramAsset) {
      // Set RAA attributes - set RAA attributes into RAA Attributes RAM attribute
		// Exclude the RAM artifacts and exclude artifacts list      
      Set<String> excludedAttributeNames = new HashSet<String>();
      excludedAttributeNames.add("DetailsURL");      // Always Skip DetailsUrl
      excludedAttributeNames.add("CanBeImpactSeed"); // Always Skip CanBeImpactSeed
      excludedAttributeNames.add("id");              // Always Skip id
      excludedAttributeNames.add("type");            // Always Skip 
      excludedAttributeNames.add("Name");            // Always Skip 
      if (raaAttributesToPopulateInRam != null ) {
         Element typeElement = raaAttributesToPopulateInRam.getChild(raaArtifact.getAttributeValue("type"));
         if (typeElement != null) {
            @SuppressWarnings("unchecked")
            List<Element> atrrs = typeElement.getChildren();
            for (Iterator iter = atrrs.iterator(); iter.hasNext();) {
               Element e = (Element) iter.next();         
               excludedAttributeNames.add(e.getName());
            }            
         }
      }
      // Build attribute name / value list with exclude list    
      addAttribute(ramAsset, DmhArtifact.getRamRaaAttributesName(), raaArtifact.getRaaAttributesFormatted(excludedAttributeNames));
      // Set the Details link attribute
      addAttribute(ramAsset, DmhArtifact.getRamRaaAttributesNameDetailsPage(), raaArtifact.getRaaDetailsPageLink());     
      addAttribute(ramAsset, DmhArtifact.getRamRaaAttributesNameRaaHome(), raaArtifact.getRaaHomepageLink());
      addAttribute(ramAsset, DmhArtifact.makeRamAttributeName("type"), raaArtifact.getAttributeValue("type"));
            
   }

   private void setRamAssetPropertiesCustomAttributes(DmhArtifact raaArtifact, RAMAsset ramAsset) {
      if (raaAttributesToPopulateInRam == null) return;
      Element typeElement = raaAttributesToPopulateInRam.getChild(raaArtifact.getAttributeValue("type"));
      if (typeElement == null) return;
      @SuppressWarnings("unchecked")
      List<Element> atrrs = typeElement.getChildren();
      for (Iterator iter = atrrs.iterator(); iter.hasNext();) {
         Element e = (Element) iter.next();         
         addAttribute(ramAsset, e.getAttributeValue("ramName"), raaArtifact.getAttributeValue(e.getName()));
      }
   }//end serRamAssetPropertiesCustomAttributes
	
   /**
    * add value for a given attribute; 
    * @param ramAsset - asset for
    * @param attributeName
    * @param attributeValue
    */
   private void addAttribute(RAMAsset ramAsset, String attributeName, String attributeValue) {
      // Note: attribute must exist before new ramAsset is created
      AssetAttribute attribute = ramAsset.getAssetAttribute( attributeName );
    
    	  
      attribute.setValues(new String[]{attributeValue});
   }//end addAttribute
	
	/** Searches RAM for RAA matching the details link to XML REST argument details.  
	 * Returns the best result as a RAM asset or null if not found
	 * @param raaDetailsUri = the details page link found in the RAA REST interface
	 * @return RAMAsset
	 */
	public RAMAsset getRamAssetByRaaUri(String raaDetailsUri, String u, String p){
		//retrieve the RAM names of the artifacts
		DmhArtifact artifact = new DmhArtifact(raaDetailsUri, u, p);
		return getRamAssetByGuid(artifact.getRamGUID(), DmhArtifact.getVersion());		
	}//end fetchbyRaaUri

	/** Searches RAM for a name matching the argument.  
    * Returns the best result as a RAM asset or null if not found
    * @param name = the name of the RAM asset to fetch
    * @return RAMAsset
    */
   public RAMAsset getRamAssetByNameString(String name){
   	RAMAsset assetFound = null;
   	RAMSession session = getRAMSession();
   
   	RAMAssetQueryBuilder qBldr = new RAMAssetQueryBuilder(session);
   	qBldr.addQueryField(qBldr.QUERY_FIELD_NAME, name);
   	SearchResult sr = session.getAssets(qBldr);
   	RAMAsset[] assets = (RAMAsset[])sr.getAssets();
   	if (assets != null && assets.length > 0) {	// found at least one
   		assetFound = assets[0];  				// return best match (should be only one - our names are unique)
   	}
   	return assetFound;
   }//end fetchByName
   
   /** Searches RAM for all RAA owned assets.  
    * @return RAMAsset
    */
   public RAMAsset[] getRaaAssets(){
   	RAMSession session = getRAMSession();
   	RAMAssetQueryBuilder qBldr = new RAMAssetQueryBuilder(session);
   	qBldr.addQueryField(qBldr.QUERY_FIELD_OWNER, "RAA");
   	SearchResult sr = session.getAssets(qBldr);
   	RAMAsset[] assets = (RAMAsset[])sr.getAssets();
   	return assets;
   }//end fetchRaaAssets

   /** Searches RAM for GUID matching the argument.  
	 * Returns the best result as a RAM asset or null if not found
	 * @param ramGuid = usually RAA assigned GUID
	 * @return RAMAsset
	 */
	public RAMAsset getRamAssetByGuid(String ramGuid, String version){
      RAMAsset assetFound = null;
		try {
         assetFound = getRAMSession().getAsset(ramGuid, version);
      } catch (com.ibm.ram.common.data.exception.RAMRuntimeException e) {
         // All okay - simply not found - we return null (the default)
      }
		return assetFound;
	} 
		
	/**
    * Fetches the 2 specified assets that have already been loaded into RAM
    * and  creates the specified relationship between them.  This also automatically
    * creates the neccessary inverse relationships.
    * @param name1 = name of the first asset
    * @param name2 = name of the second asset
    * @param relationshipName = the name of the relationship to relate name1 to name2
    */
   public void addRelationship(RAMAsset asset1, RAMAsset asset2, String relationshipName, String reverseName) {
   	//Relate the 1st asset to the second asset (create new relationship if needed)
   	asset1.addRelatedAsset(asset2,  createRelationship(relationshipName, reverseName));
   	
   	getRAMSession().queueAssetForPut(asset1);
   	getRAMSession().queueAssetForPut(asset2);
    
   	putAssets();
   }//end addRelationship
   
   /**
	 * Just a convenience method wrapping session.commitChanges(new NullProgressMonitor())
	 */
	public void putAssets() {
		try {
			getRAMSession().putAssets(new NullProgressMonitor());
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}		
	}//end commitChanges

	/** categorizes the RAMAsset as as "RAA Category/category"
	 *  String category must be a member of RAA category. so "Program" or "File"
	 *  @param asset = the RAMAsset to categorize
	 *  @param category = the subcategory to categorize the asset as.  ie "Program", "File", etc.
	 */
	private void categorizeAsset(RAMAsset asset, String category){	
		CategorySchema RAASchema = asset.getAvailableCategorySchema((DmhArtifact.getRamCategorySchemaName()));		
		RAMCategory modelCategory = (RAMCategory)RAASchema.getCategory(DmhArtifact.getRamCategoryModelName());
		
		asset.categorize(modelCategory, category);	//set the category
	 }//end categorizeAsset
	
   /**
	 * Creates a link artifact.  This method does not commitChanges to allow support
	 * for adding multiple links at once.
	 * @param asset = RAMAsset to insert the artifact into
	 * @param linkName = the name that the url will appear as in RAM
	 * @param url = the url you wish to link to
	 */
   public void insertArtifactLink(RAMAsset asset, DmhArtifact raaArtifact, String raaRelationshipName) {
	   RAMFolderArtifact root = (RAMFolderArtifact)asset.getArtifactsRoot();
      // obtain link to the corresponding details page
      String raaDetailsLink = raaArtifact.getAttributeValue("DetailsURL");
	   RAMURLArtifact link = new RAMURLArtifact(raaDetailsLink);
      String linkName = raaArtifact.getRamName() + "("+raaArtifact.getAttributeValue("id")+")";
      link.setLabel(raaRelationshipName);
      if (raaRelationshipName.equals("File")) {  // for File use full source location 
         String location = raaArtifact.getAttributeValue("SourceLocation");
         if (location != null && !location.trim().equals("")) {
            linkName = location;
         }
      }
      link.setName(linkName);
	   root.addArtifact(DmhArtifact.makeRamRelationshipName(raaRelationshipName), link);
   }//end insertArtifact

   /**
    * @param ramAsset = the asset of which to remove the artifact link
    * @param raaRelationshipName = the name (minus "RAA") of the artifact to delete
    */
   public void removeRamInsertedArtifactLinks(RAMAsset asset, String raaRelationshipName) {
	   RAMFolderArtifact root = (RAMFolderArtifact)asset.getArtifactsRoot();
	   Artifact[] links = root.getChildren();
	   for (int i = 0; i < links.length; i++) {
		   if (links[i].getName().startsWith(DmhArtifact.makeRamRelationshipName(raaRelationshipName))){
			   root.removeArtifact(links[i]);
			   getRAMSession().put(asset, new NullProgressMonitor());
		   }  
      }
   }//end removeRamInsertedArtifactLinks
   
   /**
    * given JDOM element submits the asset to RAM
    * returns true if submiton went okay
    * @param dmhDetalsLinkXml = link to the XML representation of the RAA REST interface
    * @return ramUid
    */
   public RAMAsset submitDmhArtifactToRam(String dmhDetailsLinkXml, String u, String p) {
	   //create a new DmhArtifact
	   DmhArtifact artifact = new DmhArtifact(dmhDetailsLinkXml, u, p);
	   return uploadAsset(artifact);
   }//end submitDmhArtifactToRam

   /**
	 * remove all relationships of a given relationship type from a single RAMAsset
	 * @param asset = the RAMAsset to remove the relationships from
	 * @param relationshipName = the relationship name to remove
	 */
	public void  removeRamRelationship(RAMAsset asset, String relationshipName){
		Relationship[] relationships = asset.getRelationships();
		
		for(int i = 0; i < relationships.length; i++)	{
			if(relationshipName.equals(relationships[i].getRelationshipTypeName()))	   {
				asset.removeRelationship(relationships[i]);								
            }
		}
		putAssets();
	}//end removeRelationship
		
   /**
    * @param attributesElement
    */
   @SuppressWarnings("unchecked")
   public void setRaaAttributesToPopulateInRam(Element attributesElement) {
      raaAttributesToPopulateInRam = attributesElement;
      List<Element> atrrs = raaAttributesToPopulateInRam.getChildren();
      for (Iterator iter = atrrs.iterator(); iter.hasNext();) {
         Element types = (Element) iter.next();
         // For each type - iterate through attributes
         List<Element> attributes = types.getChildren();
         for (Iterator iterator = attributes.iterator(); iterator.hasNext();) {
            Element e = (Element) iterator.next();
            String ramAttrName = e.getAttributeValue("ramName");
            if (ramAttrName == null || ramAttrName.trim().equals("")) { 
               ramAttrName = DmhArtifact.makeRamAttributeName(e.getName());
            }
            createAttribute(ramAttrName, DmhArtifact.makeRamAttributeDescription(e.getName()));                              
         }
      }
   }//end setRaaAttributesToPopulateInRam
   
   /**
    * Simple implementation of the Authenticator API, used to pass
    * user/pw info in HTTP headers.
    */
   public static class RaaAuthenticator extends Authenticator {
      String user, pw;
      
      public RaaAuthenticator(String user, String pw) {
         this.user = user;
         this.pw = pw;
      }
      
      protected PasswordAuthentication getPasswordAuthentication() {
         return new PasswordAuthentication(user, pw.toCharArray());
      }  
   } 

   /** Checks a uri and returns the responce code.  This is used to check the XML REST interface
    * to see if the artifact still exists in RAA
    * @param uri = the uri to get the responce code from
    * @return responceCode = the responce code when atempting to access uri
    */
   public int getResponceCode(String uri){
       try {
         URL url = new URL(uri);    
         Authenticator.setDefault(new RaaAuthenticator(raaUser, raaPassword));
         URLConnection con = url.openConnection();      
         if (con instanceof HttpURLConnection) {
           HttpURLConnection c = (HttpURLConnection)con;
           c.connect();
           return c.getResponseCode();
         }
       } catch (IOException e) {
         e.printStackTrace();
       }
	   return 0;
   }
   
   public void synchronizeDeleted(boolean delete){
	   RAMAsset[] raaAssets = getRaaAssets();	//fetch all RAA assets
		for (int i = 0; i < raaAssets.length; i++){		//itereate over all assets
			String link = null;
			for (int j = (raaAssets[i].getShortDescription().length()-1); j >=0; j--){	//get the REST link from the short decription
				if (raaAssets[i].getShortDescription().charAt(j) == ' '){
					link = raaAssets[i].getShortDescription().substring(j+1);
					break;
				}
			}
			
			//get the responce code and modify the asset if its corresponding artifact does not exist
			if (getResponceCode(link) == 404){
				
				if (delete == true){ //delete the asset.  do not retire
					raaAssets[i].setAction(RAMAction.DELETE);
				} else { //retire the asset
					raaAssets[i].setAction(RAMAction.RETIRE);
				}
				getRAMSession().put(raaAssets[i], new NullProgressMonitor());		
			}
		}
   }

   
   /**
    * Tests for DmhRam can be found at com.ibm.ram.test DmhRamTest.java
    */
   public static void main(String[] args) {
	   
	   
//		String repositoryUrl = "http://wsaa.svl.ibm.com:9080/com.ibm.ram.repository.web.ws.was"; //M3 on wsaa
//		String userName = "admin";
//		String password = "admin";
//		DmhRam ram = new DmhRam(repositoryUrl, userName, password);	
		
		System.out.println("fin.");
	}//end main
}//end class CreateAsset
