/* copyright-start
  ******************************************************************************
  *  Licensed Material - Property of IBM
  *
  *  5724-V27, 5655-R10, 5655-M22, 5655-L21, 5655-I49, 5655-D92
  *  Copyright IBM Corp. 1993, 2009                          
  *
  *  The source code for this program is not published or otherwise
  *  divested of its trade secrets, irrespective of what has been
  *  deposited with the U.S. Copyright Office.
  ******************************************************************************
  * copyright-end */

package com.ibm.dmh.ram;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import org.eclipse.core.runtime.NullProgressMonitor;
import org.jdom.*;
import org.jdom.filter.*;
import org.jdom.input.*;

import com.ibm.dmh.rest.client.*;
import com.ibm.ram.client.*;

/**
 * 
 * Provides methods to populate RAM with data from RAA 
 * Uses RAA REST and RAM Web Services Eclipse API toolkit  
 *
 */
public class DmhRamController {
	private boolean logging = false;

	DmhRam ramInterface = null;
	private String ramUri = null;
	private String raaUri = null;
	private String ramUser     = null;
	private String ramPassword = null;
	private String raaUser     = null;
	private String raaPassword = null;

	/**
	 * Submit to RAM all what is specified in a control file  
	 * @param controlFile
	 * @throws IOException 
	 */
	public void processAll(String controlFile) {
		File file = new File( controlFile ); // create a File object

		if (!file.exists()) {
			System.out.println("File with name \""+controlFile+"\""+" does not exist.  Please verify the location of the XML control file.");
			return;
		}

		// Populate all specified in control file into RAM
		Document config = createDocument(new File(controlFile));
		//  Store file name in DOM
		if (config == null)
			return;
		config.getRootElement().addContent((new Element("FILE")).setText(controlFile));
		processAll(config);
	}		

	/**
	 * Submit to RAM all what is specified in a control document
	 * @param controlFile
	 * @throws IOException 
	 */
	public void processAll(Document config) {
		// Populate all specified in control file into RAM
		Element r = config.getRootElement();
		readAndSetLoggingLevel(r);       // read and set logging level 
		// get RAM URI and security info
		Element ram = r.getChild("Ram"); 
		ramUri      = ram.getAttributeValue("href");
		ramUser     = ram.getAttributeValue("user");
		ramPassword = ram.getAttributeValue("password");

		ramInterface = new DmhRam(ramUri,ramUser,ramPassword);

		// get RAA URI
		Element raa = r.getChild("Raa"); 
		raaUri      = raa.getAttributeValue("href");
		raaUser     = raa.getAttributeValue("user");
		raaPassword = raa.getAttributeValue("password");
		if (logging) log("Submiting data\n"+
				"   from "+raaUri + "\n" +
				"     to "+ramUri + "User: "+ramUser +"\n");


		ramInterface.setRamEnvironment((new DmhRestClientRoot(raaUri, raaUser, raaPassword)).getSupportedTypeNames()); // Initialize the environment if not already initialized

		// Do delete if requested - if Delete Ram environment requested - this is the only thing done in response to this control file
		Element deleteEnv = r.getChild("DeleteEnvironment");
		if (deleteEnv != null) {
			log("Deleting the Integration Environment...");
			ramInterface.deleteRaaRamEnvironment();
			return; 
		}

		// Get list of RAA attributes that must be propagated as RAM attributes      
		Filter f = new ElementFilter("attributes-propagate-as-ram-asset-attributes");
		if (config.getRootElement().getDescendants(f).hasNext()) {
			Element attributesElement = (Element) config.getRootElement().getDescendants(f).next();
			ramInterface.setRaaAttributesToPopulateInRam(attributesElement);
		}

		// decide wether to delete, sync, or upload new data.
		// precedence: DeleteEnvironment > Synchronize > SubmitToRam
		Element synch = r.getChild("Synchronize");
		if (synch != null) { // we want to synchronize with REST interface
			log("Synchronizing...");
			// determine if retiring or deleting the asset
			String delete = synch.getAttributeValue("delete");
			log("delete = "+delete);
			if (delete == null || delete.equals("false")){ 
				// we want to retire the asset. do not delete it
				log(". RETIRING any inconsistent assets...\n");
				ramInterface.synchronizeDeleted(false);
			}
			else {// delete the asset. do not retire it
				log(". DELETING any inconsistent assets...\n");
				ramInterface.synchronizeDeleted(true);
			}
		}

		Element toRam = r.getChild("SubmitToRam");
		if (toRam != null ) {
			Element assetTypes = toRam.getChild("AssetTypes");
			@SuppressWarnings("unchecked")
			List<Element> types = assetTypes.getChildren();
			for (int i = 0; i < types.size(); i++) {
				String typeName = types.get(i).getName();
				System.out.println(">> "+typeName);
				submitType(typeName, types.get(i));
			}
		}
	}


	private void readAndSetLoggingLevel(Element r) {
		// Get the logging parameter
		if (r.getChild("LOG") != null && r.getChild("LOG").getTextNormalize().equals("true")) {
			logging = true; 
			log("Logging started... \n");
			if (r.getChild("FILE") != null) 
				log("       control file: "+r.getChild("FILE").getText());
		}
	}

	/**
	 * Submit to RAM all artifacts of a given type based on related control file element 
	 * @param typeName
	 * @param specs - Dom element with specifications for the submition (conditions and other control parameters)
	 * @throws IOException 
	 */
	private void submitType(String typeName, Element specs)  {
		DmhRestClientSummary summary = new DmhRestClientSummary(raaUri, typeName, raaUser, raaPassword);
		// Iterate through all pages of all documents of a given type
		while (summary.nextPage()) {
			List<Element> pgms = summary.getPageArtifactElements();
			for (Iterator iter = pgms.iterator(); iter.hasNext();) {
				Element e = (Element) iter.next();
				// for each artifact encountered - check conditions if any and populate RAM 
				if (logging) log(e.getText()+"="+e.getAttributeValue("href")+"\n");
				String detailsLink = e.getAttributeValue("href");
				if (checkConditions(detailsLink, specs)) { // ? conditions for submition met
					RAMAsset ramAsset = ramInterface.submitDmhArtifactToRam(detailsLink, raaUser, raaPassword); // submit as asset to ram
					submitRelatedAssets(ramAsset, detailsLink, specs);
					submitChildArtifacts(ramAsset, detailsLink, specs);
				}
			}
		}
	}

	/**
	 * 
	 * @param detailsLink AA xml REST details link
	 * @param specs JDOM element with conditions to be met for a given asset
	 * @return true if conditoons met or there are no conditions specified
	 */
	private boolean checkConditions(String detailsLink, Element specs) {
		/* The element passed is main element for type, sample XML: 
		 *  <Program>
		 *       <conditions>
		 *       	<attributes description="Attribute name and values that must be matched to submit to ram">
		 *       		<Name>QAD02</Name>
		 *       	</attributes>
		 *       </conditions>
		 */
		boolean conditionMet = true;
		if (specs != null && specs.getChild("conditions") != null && specs.getChild("conditions").getChild("attributes") != null) {
			@SuppressWarnings("unchecked")
			List<Element> attribs = specs.getChild("conditions").getChild("attributes").getChildren();
			for (Iterator<Element> iter = attribs.iterator(); iter.hasNext();) {
				Element a = iter.next();
				String attributeName = a.getName();
				String attributeValueCondition = a.getTextNormalize();
				String attributeValueActual = DmhRestClientDetails.getAttribute(attributeName, detailsLink, raaUser, raaPassword);
				if (logging) log("Condition "+attributeName+"="+attributeValueCondition+" actual:"+attributeValueActual);

				//Compile regular expression
				Pattern p = Pattern.compile(attributeValueCondition);

				if (! p.matcher(attributeValueActual).find()) {
					conditionMet = false;
					break;
				}
			}

		}
		if (logging) log(" - "+conditionMet+"\n");
		return conditionMet;
	}

	/**
	 * Submits to RAM artifacts from seleced relationships and creates 
	 * correspinding relationship in RAM
	 * @param detailsLink
	 * @param specs control file type top level element specification for relationsips creation ( ex: <Program> ) 
	 * @throws IOException 
	 */
	private void submitRelatedAssets(RAMAsset ramAsset, String detailsLink, Element specs)  {
		/* 
		 * <Program>
		 *      ... 
		 *	<!-- Only selected relationships will be submited to RAM -->
		 *  	<asset-relationships description="Relationships to submit to RAM, note: all related artifacts will also be submited as assets.">
		 *  		<FilesNeededToBuild/>
		 *      . . . 
		 */
		if (specs != null && specs.getChild("asset-relationships") != null ) {
			@SuppressWarnings("unchecked")
			List<Element> ralats = specs.getChild("asset-relationships").getChildren();
			// Iterate over all listed relationships
			for (Iterator<Element> iter = ralats.iterator(); iter.hasNext();) {
				Element r = iter.next();                // point to relationship element in control file  
				String relationshipName = r.getName();  // relationship name from control element
				String reverseRelationshipName = r.getAttributeValue("reverseName");
				ramInterface.removeRamRelationship(ramAsset, relationshipName); // remove old relationships if any
				if (logging) log("submitRelatedAssets "+relationshipName+"\n");
				DmhRestClientRelationship relatedLinks = new DmhRestClientRelationship(detailsLink, r.getName(), raaUser, raaPassword);
				// For each relationship listed retrieve all pages of related artifact links  
				while (relatedLinks.nextPage()) {
					List<Element> linksPage = relatedLinks.getPageArtifactElements();
					// For each page of atrifatc links - create artifact (if not already exist) 
					// and the realtionship  
					for (Iterator iPage = linksPage.iterator(); iPage.hasNext();) {
						Element e = (Element) iPage.next();
						String relatedRestLink = e.getAttributeValue("href");
						if(logging) log ("    creating related artifact: "+relatedRestLink+"\n");
						RAMAsset a = ramInterface.getRamAssetByRaaUri(relatedRestLink, raaUser, raaPassword);
						RAMAsset relatedAsset = null;
						if (a == null) {  // if not found - create a new asset
							relatedAsset = ramInterface.submitDmhArtifactToRam(relatedRestLink, raaUser, raaPassword);
						} else { // asset found in RAM   
							relatedAsset = a;
						}
						if(logging) log ("    creating relationship: "+relationshipName+" from "+detailsLink+" to "+ relatedRestLink + "\n");
						ramInterface.addRelationship(ramAsset, relatedAsset, relationshipName, reverseRelationshipName);
					}			
				}
			}		
		}
	}

	/**
	 * Adds artifacts from selected relationships as part of the paren RAM asset 
	 * @param detailsLink
	 * @param specs
	 */
	private void submitChildArtifacts(RAMAsset ramAsset, String detailsLink, Element specs) {
		/*
		 * <Program>
		 *      ... 
		 *	<!-- Only selected relationships will be submited to RAM -->
		 *   	<containment-relationships description="Relationships to submit to RAM, note: related artifacts will be submited as contained artifacts.">
		 *   		<DataElementsTopLevel/>
		 */
		if (specs != null && specs.getChild("containment-relationships") != null ) {
			@SuppressWarnings("unchecked")
			List<Element> ralats = specs.getChild("containment-relationships").getChildren();
			//  Iterate through all specified relationships
			for (Iterator<Element> iter = ralats.iterator(); iter.hasNext();) {
				Element r = iter.next(); // Get relationship element
				String raaRelationshipName = r.getName(); // relationship name
				ramInterface.removeRamInsertedArtifactLinks(ramAsset, raaRelationshipName);
				if (logging) log("submitChildArtifacts "+raaRelationshipName+" into "+ detailsLink+": <");
				DmhRestClientRelationship relatedLinks = new DmhRestClientRelationship(detailsLink, raaRelationshipName, raaUser, raaPassword);
				// For each relationship listed retrieve all pages of related artifact links  
				while (relatedLinks.nextPage()) {
					List<Element> linksPage = relatedLinks.getPageArtifactElements();
					// For each page of atrifatc links - create artifact (if not already exist) 
					// and the realtionship  
					for (Iterator iPage = linksPage.iterator(); iPage.hasNext();) {
						Element e = (Element) iPage.next();
						// get link to xml details page
						String relatedRestRestLink = e.getAttributeValue("href");
						// instantiate interface to details in RAA
						DmhArtifact related = new DmhArtifact(relatedRestRestLink, raaUser, raaPassword);
						// Store lin to RAA details page as content in RAM asset
						ramInterface.insertArtifactLink(ramAsset, related, raaRelationshipName);
						if(logging) log (".");
					}			
				}
				if(logging) log (">\n");
				ramInterface.getRAMSession().put(ramAsset, new NullProgressMonitor());
			}		
		}

	}


	/**
	 * Given a file with integraton control parametrs - populate RAM   
	 * @param file
	 * @return
	 */
	public static Document createDocument(File file) {
		Document d = null;
		// This builds a document of whatever's in the given resource
		SAXBuilder builder = new SAXBuilder();
		try {
			d = builder.build(file);
		} catch (JDOMException e) {
			System.out.println("File "+file.getName()+" is not of the correct format.  For details regarding the correct format of an XML control file, please refer to the application's documentation.");
		} catch (IOException e) {;}
		return d;
	}

	protected void log(String l) {
		if (logging) System.out.print(l);
	}

	/**
	 * Tests for DmhRamController can be found at com.ibm.ram.test DmhRamControllerTest.java
	 */
	public static void main(String[] args) {
		String controlFile = null;
		System.out.println("IBM Rational Asset Analyzer 5.5\nIBM Rational Asset Manager 7.1 Integration\n");

		if (args != null && args.length > 0) {
			controlFile = args[0];
		}
		else { //accept user input for a path to the desired control file
			BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in), 1);
			System.out.print("Please enter the path the the XML control file: ");
			try {
				String temp = stdin.readLine();  //need this temporary variable here.  using "controlFile" here seems to cause the thread to continue without waiting for keyboard input.
				if(temp.startsWith("\"") && temp.endsWith("\"")) //remove quotes
					controlFile = temp.substring(1, temp.length()-1);
				else
					controlFile = temp;
				System.out.println("DmhRamController Starting ...");
				DmhRamController ctrl = new DmhRamController(); 
				ctrl.processAll(controlFile);
				System.out.println("... DmhRamControler Done");
				return;
			}
			catch(IOException e) {e.printStackTrace();}
		}
		System.out.println("DmhRamControler Starting ...");
		DmhRamController ctrl = new DmhRamController(); 
		ctrl.processAll(controlFile);
		System.out.println("... DmhRamControler Done");

	}

}
